{% extends 'wildewidgets/header_with_controls.html' %}

{% block controls %}
<div class="controls">
  <button type="button" class="btn btn-{{button_class}}" data-bs-toggle="modal" data-bs-target="#{{modal_id}}">{{button_text}}</button>
</div>  
{% endblock %}  
