{% extends 'wildewidgets/block.html' %}
{% load wildewidgets %}

{% block block_content %}
{% wildewidgets header %}
<div class="row">
    {% wildewidgets sidebar %}
    {% wildewidgets main %}
</div>
{% for modal in modals %}
    {% wildewidgets modal %}
{% endfor %}
{% endblock %}
